@@ -5,6 +5,7 @@
 #include "pgrn-global.h"
 #include "pgrn-groonga.h"
 #include "pgrn-index-status.h"
+#include "pgrn-pg.h"
 #include "pgrn-wal.h"
 #include "pgrn-writable.h"
 
@@ -128,53 +129,6 @@ struct PGrnWALData_
 };
 
 #ifdef PGRN_SUPPORT_WAL
-#	if MSGPACK_VERSION_MAJOR == 0
-#		define MSGPACK_OBJECT_FLOAT MSGPACK_OBJECT_DOUBLE
-#		define MSGPACK_OBJECT_STR   MSGPACK_OBJECT_RAW
-#		define MSGPACK_OBJECT_BIN   MSGPACK_OBJECT_RAW
-#	endif
-
-#	if MSGPACK_VERSION_MAJOR == 0
-#		define MSGPACK_OBJECT_VIA_FLOAT(object) ((object).via.dec)
-#		define MSGPACK_OBJECT_VIA_STR(object)   ((object).via.raw)
-#		define MSGPACK_OBJECT_VIA_BIN(object)   ((object).via.raw)
-#	else
-#		define MSGPACK_OBJECT_VIA_FLOAT(object) ((object).via.f64)
-#		define MSGPACK_OBJECT_VIA_STR(object)   ((object).via.str)
-#		define MSGPACK_OBJECT_VIA_BIN(object)   ((object).via.bin)
-#	endif
-#endif
-
-#ifdef PGRN_SUPPORT_WAL
-#	if MSGPACK_VERSION_MAJOR == 0
-#		define msgpack_pack_str(packer, size) \
-	msgpack_pack_raw((packer), (size))
-#		define msgpack_pack_str_body(packer, data, size) \
-	msgpack_pack_raw_body((packer), (data), (size))
-#		define msgpack_pack_bin(packer, size) \
-	msgpack_pack_raw((packer), (size))
-#		define msgpack_pack_bin_body(packer, data, size) \
-	msgpack_pack_raw_body((packer), (data), (size))
-#	endif
-
-#ifdef PGRN_SUPPORT_WAL
-#	if MSGPACK_VERSION_MAJOR == 0
-#		define MSGPACK_UNPACKER_NEXT(unpacker, unpacked) \
-	msgpack_unpacker_next((unpacker), (unpacked))
-#	else
-#		define MSGPACK_UNPACKER_NEXT(unpacker, unpacked) \
-	msgpack_unpacker_next((unpacker), (unpacked)) == MSGPACK_UNPACK_SUCCESS
-#	endif
-#endif
-
-#ifdef PGRN_SUPPORT_WAL
-#	if MSGPACK_VERSION_MAJOR == 0
-#		define MSGPACK_PACKER_WRITE_LENGTH_TYPE unsigned int
-#	else
-#		define MSGPACK_PACKER_WRITE_LENGTH_TYPE size_t
-#	endif
-#endif
-
 static void
 msgpack_pack_cstr(msgpack_packer *packer, const char *string)
 {
@@ -455,7 +409,7 @@ PGrnWALPageWriterEnsureCurrent(PGrnWALData *data)
 static int
 PGrnWALPageWriter(void *userData,
 				  const char *buffer,
-				  MSGPACK_PACKER_WRITE_LENGTH_TYPE length)
+				  size_t length)
 {
 	PGrnWALData *data = userData;
 	int written = 0;
@@ -482,7 +436,10 @@ PGrnWALPageWriter(void *userData,
 		}
 
 		if (PGrnWALPageGetFreeSize(data->current.page) == 0)
+		{
 			PGrnWALPageFilled(data);
+			PGrnWALPageWriterEnsureCurrent(data);
+		}
 	}
 
 	return written;
@@ -1114,9 +1071,6 @@ PGrnWALApplyNeeded(PGrnWALApplyData *data)
 	LocationIndex currentOffset;
 	BlockNumber nBlocks;
 
-	if (!PGrnWALEnabled)
-		return false;
-
 	PGrnIndexStatusGetWALAppliedPosition(data->index,
 										 &currentBlock,
 										 &currentOffset);
@@ -1128,33 +1082,46 @@ PGrnWALApplyNeeded(PGrnWALApplyData *data)
 	{
 		return false;
 	}
-	else if (currentBlock == (nBlocks - 1))
-	{
-		Buffer buffer;
-		Page page;
-		bool needToApply;
-
-		buffer = PGrnWALReadLockedBuffer(data->index,
-										 currentBlock,
-										 BUFFER_LOCK_SHARE);
-		page = BufferGetPage(buffer);
-		needToApply = (PGrnWALPageGetLastOffset(page) > currentOffset);
-		UnlockReleaseBuffer(buffer);
-		if (!needToApply)
-			return false;
-	}
 	else
 	{
 		Buffer buffer;
 		Page page;
+		size_t freeSize;
+		LocationIndex offset;
+		bool haveDataInCurrentPage;
 		bool needToApply;
 
 		buffer = PGrnWALReadLockedBuffer(data->index,
 										 currentBlock,
 										 BUFFER_LOCK_SHARE);
 		page = BufferGetPage(buffer);
-		needToApply = (PGrnWALPageGetLastOffset(page) > 0);
+		freeSize = PGrnWALPageGetFreeSize(page);
+		offset = PGrnWALPageGetLastOffset(page);
 		UnlockReleaseBuffer(buffer);
+		haveDataInCurrentPage = (offset > currentOffset);
+		if (haveDataInCurrentPage)
+		{
+			needToApply = true;
+		}
+		else
+		{
+			BlockNumber nextBlock;
+			Buffer nextBuffer;
+			Page nextPage;
+
+			nextBlock = currentBlock + 1;
+			if (nextBlock == nBlocks)
+			{
+				/* 0 is the meta page. 1 is the first page that has data. */
+				nextBlock = 1;
+			}
+			nextBuffer = PGrnWALReadLockedBuffer(data->index,
+												 nextBlock,
+												 BUFFER_LOCK_SHARE);
+			nextPage = BufferGetPage(nextBuffer);
+			needToApply = (PGrnWALPageGetLastOffset(nextPage) > 0);
+			UnlockReleaseBuffer(nextBuffer);
+		}
 		if (!needToApply)
 			return false;
 	}
@@ -1184,9 +1151,9 @@ PGrnWALApplyKeyEqual(PGrnWALApplyData *data,
 	}
 
 	nameSize = strlen(name);
-	if (MSGPACK_OBJECT_VIA_STR(*key).size != nameSize)
+	if (key->via.str.size != nameSize)
 		return false;
-	if (memcmp(MSGPACK_OBJECT_VIA_STR(*key).ptr, name, nameSize) != 0)
+	if (memcmp(key->via.str.ptr, name, nameSize) != 0)
 		return false;
 
 	return true;
@@ -1207,8 +1174,8 @@ PGrnWALApplyValueGetPositiveInteger(PGrnWALApplyData *data,
 					RelationGetRelationName(data->index),
 					RelationGetRelid(data->index),
 					context ? " " : "",
-					MSGPACK_OBJECT_VIA_STR(kv->key).size,
-					MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
+					kv->key.via.str.size,
+					kv->key.via.str.ptr,
 					kv->val.type);
 	}
 
@@ -1232,13 +1199,13 @@ PGrnWALApplyValueGetString(PGrnWALApplyData *data,
 					RelationGetRelationName(data->index),
 					RelationGetRelid(data->index),
 					context ? " " : "",
-					MSGPACK_OBJECT_VIA_STR(kv->key).size,
-					MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
+					kv->key.via.str.size,
+					kv->key.via.str.ptr,
 					kv->val.type);
 	}
 
-	*string = MSGPACK_OBJECT_VIA_STR(kv->val).ptr;
-	*stringSize = MSGPACK_OBJECT_VIA_STR(kv->val).size;
+	*string = kv->val.via.str.ptr;
+	*stringSize = kv->val.via.str.size;
 }
 
 static void
@@ -1258,13 +1225,13 @@ PGrnWALApplyValueGetBinary(PGrnWALApplyData *data,
 					RelationGetRelationName(data->index),
 					RelationGetRelid(data->index),
 					context ? " " : "",
-					MSGPACK_OBJECT_VIA_BIN(kv->key).size,
-					MSGPACK_OBJECT_VIA_BIN(kv->key).ptr,
+					kv->key.via.bin.size,
+					kv->key.via.bin.ptr,
 					kv->val.type);
 	}
 
-	*binary = MSGPACK_OBJECT_VIA_BIN(kv->val).ptr;
-	*binarySize = MSGPACK_OBJECT_VIA_BIN(kv->val).size;
+	*binary = kv->val.via.bin.ptr;
+	*binarySize = kv->val.via.bin.size;
 }
 
 static grn_obj *
@@ -1281,8 +1248,8 @@ PGrnWALApplyValueGetGroongaObject(PGrnWALApplyData *data,
 		object = NULL;
 		break;
 	case MSGPACK_OBJECT_STR:
-		object = PGrnLookupWithSize(MSGPACK_OBJECT_VIA_STR(kv->val).ptr,
-									MSGPACK_OBJECT_VIA_STR(kv->val).size,
+		object = PGrnLookupWithSize(kv->val.via.str.ptr,
+									kv->val.via.str.size,
 									ERROR);
 		break;
 	default:
@@ -1293,8 +1260,8 @@ PGrnWALApplyValueGetGroongaObject(PGrnWALApplyData *data,
 					RelationGetRelationName(data->index),
 					RelationGetRelid(data->index),
 					context ? " " : "",
-					MSGPACK_OBJECT_VIA_STR(kv->key).size,
-					MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
+					kv->key.via.str.size,
+					kv->key.via.str.ptr,
 					kv->val.type);
 		break;
 	}
@@ -1321,8 +1288,8 @@ PGrnWALApplyValueGetGroongaObjectIDs(PGrnWALApplyData *data,
 					RelationGetRelationName(data->index),
 					RelationGetRelid(data->index),
 					context ? " " : "",
-					MSGPACK_OBJECT_VIA_STR(kv->key).size,
-					MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
+					kv->key.via.str.size,
+					kv->key.via.str.ptr,
 					kv->val.type);
 	}
 
@@ -1344,14 +1311,14 @@ PGrnWALApplyValueGetGroongaObjectIDs(PGrnWALApplyData *data,
 						RelationGetRelationName(data->index),
 						RelationGetRelid(data->index),
 						context ? " " : "",
-						MSGPACK_OBJECT_VIA_STR(kv->key).size,
-						MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
+						kv->key.via.str.size,
+						kv->key.via.str.ptr,
 						i,
 						element->type);
 		}
 
-		object = PGrnLookupWithSize(MSGPACK_OBJECT_VIA_STR(*element).ptr,
-									MSGPACK_OBJECT_VIA_STR(*element).size,
+		object = PGrnLookupWithSize(element->via.str.ptr,
+									element->via.str.size,
 									ERROR);
 		objectID = grn_obj_id(ctx, object);
 		GRN_RECORD_PUT(ctx, ids, objectID);
@@ -1376,8 +1343,8 @@ PGrnWALApplyValueGetTableModule(PGrnWALApplyData *data,
 		GRN_BULK_REWIND(module);
 		GRN_TEXT_SET(ctx,
 					 module,
-					 MSGPACK_OBJECT_VIA_STR(kv->val).ptr,
-					 MSGPACK_OBJECT_VIA_STR(kv->val).size);
+					 kv->val.via.str.ptr,
+					 kv->val.via.str.size);
 		break;
 	default:
 		PGrnCheckRC(GRN_INVALID_ARGUMENT,
@@ -1387,8 +1354,8 @@ PGrnWALApplyValueGetTableModule(PGrnWALApplyData *data,
 					RelationGetRelationName(data->index),
 					RelationGetRelid(data->index),
 					context ? " " : "",
-					MSGPACK_OBJECT_VIA_STR(kv->key).size,
-					MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
+					kv->key.via.str.size,
+					kv->key.via.str.ptr,
 					kv->val.type);
 		break;
 	}
@@ -1414,8 +1381,8 @@ PGrnWALApplyValueGetTableModules(PGrnWALApplyData *data,
 		GRN_BULK_REWIND(modules);
 		GRN_TEXT_SET(ctx,
 					 modules,
-					 MSGPACK_OBJECT_VIA_STR(kv->val).ptr,
-					 MSGPACK_OBJECT_VIA_STR(kv->val).size);
+					 kv->val.via.str.ptr,
+					 kv->val.via.str.size);
 		break;
 	case MSGPACK_OBJECT_ARRAY:
 	{
@@ -1440,8 +1407,8 @@ PGrnWALApplyValueGetTableModules(PGrnWALApplyData *data,
 							RelationGetRelationName(data->index),
 							RelationGetRelid(data->index),
 							context ? " " : "",
-							MSGPACK_OBJECT_VIA_STR(kv->key).size,
-							MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
+							kv->key.via.str.size,
+							kv->key.via.str.ptr,
 							i,
 							element->type);
 			}
@@ -1450,8 +1417,8 @@ PGrnWALApplyValueGetTableModules(PGrnWALApplyData *data,
 				GRN_TEXT_PUTS(ctx, modules, ", ");
 			GRN_TEXT_PUT(ctx,
 						 modules,
-						 MSGPACK_OBJECT_VIA_STR(*element).ptr,
-						 MSGPACK_OBJECT_VIA_STR(*element).size);
+						 element->via.str.ptr,
+						 element->via.str.size);
 		}
 		break;
 	}
@@ -1463,8 +1430,8 @@ PGrnWALApplyValueGetTableModules(PGrnWALApplyData *data,
 					RelationGetRelationName(data->index),
 					RelationGetRelid(data->index),
 					context ? " " : "",
-					MSGPACK_OBJECT_VIA_STR(kv->key).size,
-					MSGPACK_OBJECT_VIA_STR(kv->key).ptr,
+					kv->key.via.str.size,
+					kv->key.via.str.ptr,
 					kv->val.type);
 		break;
 	}
@@ -1557,13 +1524,13 @@ PGrnWALApplyInsertArray(PGrnWALApplyData *data,
 			break;
 #undef ELEMENT_VALUE
 		case MSGPACK_OBJECT_FLOAT:
-			GRN_FLOAT_PUT(ctx, value, MSGPACK_OBJECT_VIA_FLOAT(*element));
+			GRN_FLOAT_PUT(ctx, value, element->via.f64);
 			break;
 		case MSGPACK_OBJECT_STR:
 			grn_vector_add_element(ctx,
 								   value,
-								   MSGPACK_OBJECT_VIA_STR(*element).ptr,
-								   MSGPACK_OBJECT_VIA_STR(*element).size,
+								   element->via.str.ptr,
+								   element->via.str.size,
 								   0,
 								   element_domain_id);
 			break;
@@ -1645,8 +1612,8 @@ PGrnWALApplyInsert(PGrnWALApplyData *data,
 		}
 
 		column = PGrnLookupColumnWithSize(table,
-										  MSGPACK_OBJECT_VIA_STR(*key).ptr,
-										  MSGPACK_OBJECT_VIA_STR(*key).size,
+										  key->via.str.ptr,
+										  key->via.str.size,
 										  ERROR);
 		switch (value->type)
 		{
@@ -1664,13 +1631,13 @@ PGrnWALApplyInsert(PGrnWALApplyData *data,
 			break;
 		case MSGPACK_OBJECT_FLOAT:
 			grn_obj_reinit(ctx, walValue, GRN_DB_FLOAT, 0);
-			GRN_FLOAT_SET(ctx, walValue, MSGPACK_OBJECT_VIA_FLOAT(*value));
+			GRN_FLOAT_SET(ctx, walValue, value->via.f64);
 			break;
 		case MSGPACK_OBJECT_STR:
 			grn_obj_reinit(ctx, walValue, GRN_DB_TEXT, 0);
 			GRN_TEXT_SET(ctx, walValue,
-						 MSGPACK_OBJECT_VIA_STR(*value).ptr,
-						 MSGPACK_OBJECT_VIA_STR(*value).size);
+						 value->via.str.ptr,
+						 value->via.str.size);
 			break;
 		case MSGPACK_OBJECT_ARRAY:
 			PGrnWALApplyInsertArray(data,
@@ -1989,7 +1956,7 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 						message,
 						currentBlock,
 						currentOffset,
-						MSGPACK_OBJECT_VIA_FLOAT(*object));
+						object->via.f64);
 			break;
 		case MSGPACK_OBJECT_STR:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
@@ -2000,8 +1967,8 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 						message,
 						currentBlock,
 						currentOffset,
-						(int) MSGPACK_OBJECT_VIA_STR(*object).size,
-						MSGPACK_OBJECT_VIA_STR(*object).ptr);
+						(int) object->via.str.size,
+						object->via.str.ptr);
 			break;
 		case MSGPACK_OBJECT_ARRAY:
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
@@ -2024,7 +1991,7 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 						message,
 						currentBlock,
 						currentOffset,
-						MSGPACK_OBJECT_VIA_BIN(*object).size);
+						object->via.bin.size);
 			break;
 #endif
 		default:
@@ -2101,6 +2068,7 @@ PGrnWALApplyConsume(PGrnWALApplyData *data)
 	BlockNumber maxBlock;
 	msgpack_unpacker unpacker;
 	msgpack_unpacked unpacked;
+	size_t bufferedSize = 0;
 
 	msgpack_unpacker_init(&unpacker, MSGPACK_UNPACKER_INIT_BUFFER_SIZE);
 	msgpack_unpacked_init(&unpacked);
@@ -2118,62 +2086,111 @@ PGrnWALApplyConsume(PGrnWALApplyData *data)
 	maxBlock = meta->max;
 	if (maxBlock == PGRN_WAL_META_PAGE_BLOCK_NUMBER)
 		maxBlock = nBlocks;
-	for (i = 0; i < nBlocks; i++)
+	PG_TRY();
 	{
-		BlockNumber block;
-		Buffer buffer;
-		Page page;
-		LocationIndex lastOffset;
-		size_t dataSize;
-
-		block = (startBlock + i) % nBlocks;
-		if (block == PGRN_WAL_META_PAGE_BLOCK_NUMBER)
-			continue;
-		if (block > maxBlock)
-			continue;
-
-		buffer = PGrnWALReadLockedBuffer(data->index, block, BUFFER_LOCK_SHARE);
-		page = BufferGetPage(buffer);
-		lastOffset = PGrnWALPageGetLastOffset(page);
-		if (dataOffset > lastOffset)
+		for (i = 0; i < nBlocks; i++)
 		{
-			PGrnCheckRC(GRN_UNKNOWN_ERROR,
-						"[wal][apply][consume][%s(%u)] "
-						"unconsumed WAL are overwritten: "
-						"pgroonga.max_wal_size should be increased or "
-						"pgroonga_wal_applier.naptime should be decreased",
-						RelationGetRelationName(data->index),
-						RelationGetRelid(data->index));
+			BlockNumber block;
+			Buffer buffer;
+			Page page;
+			LocationIndex lastOffset;
+			size_t dataSize;
+			size_t parsedSize;
+
+			block = (startBlock + i) % nBlocks;
+			if (block == PGRN_WAL_META_PAGE_BLOCK_NUMBER)
+				continue;
+			if (block > maxBlock)
+				continue;
+
+			buffer = PGrnWALReadLockedBuffer(data->index,
+											 block,
+											 BUFFER_LOCK_SHARE);
+			page = BufferGetPage(buffer);
+			lastOffset = PGrnWALPageGetLastOffset(page);
+			if (dataOffset > lastOffset)
+			{
+				PGrnCheckRC(GRN_UNKNOWN_ERROR,
+							"[wal][apply][consume][%s(%u)]"
+							"[%u/%u/%" PGRN_PRIuSIZE "] "
+							"unconsumed WAL are overwritten: "
+							"pgroonga.max_wal_size should be increased or "
+							"pgroonga_wal_applier.naptime should be decreased",
+							RelationGetRelationName(data->index),
+							RelationGetRelid(data->index),
+							block,
+							dataOffset,
+							dataSize);
+			}
+			dataSize = lastOffset - dataOffset;
+			if (!msgpack_unpacker_reserve_buffer(&unpacker, dataSize))
+			{
+				PGrnCheckRC(GRN_NO_MEMORY_AVAILABLE,
+							"[wal][apply][consume][%s(%u)]"
+							"[%u/%u/%" PGRN_PRIuSIZE "] "
+							"failed to allocate buffer to unpack msgpack data",
+							RelationGetRelationName(data->index),
+							RelationGetRelid(data->index),
+							block,
+							dataOffset,
+							dataSize);
+			}
+			memcpy(msgpack_unpacker_buffer(&unpacker),
+				   PGrnWALPageGetData(page) + dataOffset,
+				   dataSize);
+			bufferedSize += dataSize;
+			UnlockReleaseBuffer(buffer);
+
+			msgpack_unpacker_buffer_consumed(&unpacker, dataSize);
+			while (true)
+			{
+				msgpack_unpack_return unpackResult =
+					msgpack_unpacker_next_with_size(&unpacker,
+													&unpacked,
+													&parsedSize);
+				LocationIndex appliedOffset;
+
+				if (unpackResult < 0)
+				{
+					PGrnCheckRC(GRN_UNKNOWN_ERROR,
+								"[wal][apply][consume][%s(%u)]"
+								"[%u/%u/%" PGRN_PRIuSIZE "] "
+								"failed to unpack WAL: %d: ",
+								RelationGetRelationName(data->index),
+								RelationGetRelid(data->index),
+								block,
+								dataOffset,
+								dataSize,
+								unpackResult);
+				}
+				if (unpackResult != MSGPACK_UNPACK_SUCCESS)
+				{
+					break;
+				}
+
+				PGrnWALApplyObject(data, &unpacked.data);
+				bufferedSize -= parsedSize;
+				appliedOffset = dataOffset + dataSize - bufferedSize;
+				PGrnIndexStatusSetWALAppliedPosition(data->index,
+													 block,
+													 appliedOffset);
+				nAppliedOperations++;
+			}
+
+			if (block == nextBlock)
+				break;
+
+			dataOffset = 0;
 		}
-		dataSize = lastOffset - dataOffset;
-		msgpack_unpacker_reserve_buffer(&unpacker, dataSize);
-		memcpy(msgpack_unpacker_buffer(&unpacker),
-			   PGrnWALPageGetData(page) + dataOffset,
-			   dataSize);
-		UnlockReleaseBuffer(buffer);
-		if (dataSize == 0)
-			break;
-
-		msgpack_unpacker_buffer_consumed(&unpacker, dataSize);
-		while (MSGPACK_UNPACKER_NEXT(&unpacker, &unpacked))
-		{
-			LocationIndex appliedOffset;
-
-			PGrnWALApplyObject(data, &unpacked.data);
-			appliedOffset =
-				dataOffset +
-				dataSize - (unpacker.used - unpacker.off);
-			PGrnIndexStatusSetWALAppliedPosition(data->index,
-												 block,
-												 appliedOffset);
-			nAppliedOperations++;
-		}
-
-		if (block == nextBlock)
-			break;
-
-		dataOffset = 0;
 	}
+	PG_CATCH();
+	{
+		UnlockReleaseBuffer(metaBuffer);
+		msgpack_unpacked_destroy(&unpacked);
+		msgpack_unpacker_destroy(&unpacker);
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
 	UnlockReleaseBuffer(metaBuffer);
 	msgpack_unpacked_destroy(&unpacked);
 	msgpack_unpacker_destroy(&unpacker);
@@ -2189,7 +2206,11 @@ PGrnWALApply(Relation index)
 #ifdef PGRN_SUPPORT_WAL
 	PGrnWALApplyData data;
 
+	if (!PGrnWALEnabled)
+		return 0;
+
 	data.index = index;
+
 	if (!PGrnWALApplyNeeded(&data))
 		return 0;
 
@@ -2248,7 +2269,10 @@ pgroonga_wal_apply_index(PG_FUNCTION_ARGS)
 						tag,
 						DatumGetCString(indexNameDatum));
 		}
-		nAppliedOperations = PGrnWALApply(index);
+		if (!PGRN_RELKIND_HAS_PARTITIONS(index->rd_rel->relkind))
+		{
+			nAppliedOperations = PGrnWALApply(index);
+		}
 	}
 	PG_CATCH();
 	{
@@ -2731,6 +2755,13 @@ pgroonga_wal_set_applied_position_index(PG_FUNCTION_ARGS)
 						tag,
 						DatumGetCString(indexNameDatum));
 		}
+		if (PGrnPGIsParentIndex(index))
+		{
+			PGrnCheckRC(GRN_INVALID_ARGUMENT,
+						"%s parent index for declarative partitioning: <%s>",
+						tag,
+						DatumGetCString(indexNameDatum));
+		}
 		PGrnWALLock(index);
 		PGrnIndexStatusSetWALAppliedPosition(index, block, offset);
 		PGrnWALUnlock(index);
@@ -2796,6 +2827,13 @@ pgroonga_wal_set_applied_position_index_last(PG_FUNCTION_ARGS)
 						tag,
 						DatumGetCString(indexNameDatum));
 		}
+		if (PGrnPGIsParentIndex(index))
+		{
+			PGrnCheckRC(GRN_INVALID_ARGUMENT,
+						"%s parent index for declarative partitioning: <%s>",
+						tag,
+						DatumGetCString(indexNameDatum));
+		}
 		PGrnWALLock(index);
 		PGrnWALGetLastPosition(index, &block, &offset);
 		PGrnIndexStatusSetWALAppliedPosition(index, block, offset);
@@ -2923,6 +2961,11 @@ pgroonga_wal_set_applied_position_all_last(PG_FUNCTION_ARGS)
 			RelationClose(index);
 			continue;
 		}
+		if (PGrnPGIsParentIndex(index))
+		{
+			RelationClose(index);
+			continue;
+		}
 
 		PG_TRY();
 		{
