@@ -14,6 +14,7 @@
 #	include <access/htup_details.h>
 #endif
 #include <access/xact.h>
+#include <access/xlog.h>
 #include <catalog/pg_database.h>
 #include <executor/spi.h>
 #include <miscadmin.h>
@@ -41,6 +42,8 @@ PG_MODULE_MAGIC;
 
 extern PGDLLEXPORT void _PG_init(void);
 extern PGDLLEXPORT void
+pgroonga_crash_safer_reset_position_one(Datum datum) pg_attribute_noreturn();
+extern PGDLLEXPORT void
 pgroonga_crash_safer_reindex_one(Datum datum) pg_attribute_noreturn();
 extern PGDLLEXPORT void
 pgroonga_crash_safer_flush_one(Datum datum) pg_attribute_noreturn();
@@ -119,18 +122,90 @@ pgroonga_crash_safer_sigusr1(SIGNAL_ARGS)
 }
 
 static void
-pgroonga_crash_safer_reindex_one_on_exit(int code, Datum databaseInfoDatum)
+pgroonga_crash_safer_prepare_one_on_exit(int code, Datum databaseInfoDatum)
 {
 	uint64 databaseInfo = DatumGetUInt64(databaseInfoDatum);
 	Oid databaseOid;
 	Oid tableSpaceOid;
 	PGRN_DATABASE_INFO_UNPACK(databaseInfo, databaseOid, tableSpaceOid);
-	pgrn_crash_safer_statuses_set_reindex_pid(NULL,
+	pgrn_crash_safer_statuses_set_prepare_pid(NULL,
 											  databaseOid,
 											  tableSpaceOid,
 											  InvalidPid);
 }
 
+void
+pgroonga_crash_safer_reset_position_one(Datum databaseInfoDatum)
+{
+	uint64 databaseInfo = DatumGetUInt64(databaseInfoDatum);
+	Oid databaseOid;
+	Oid tableSpaceOid;
+
+	PGRN_DATABASE_INFO_UNPACK(databaseInfo, databaseOid, tableSpaceOid);
+
+	BackgroundWorkerInitializeConnectionByOid(databaseOid, InvalidOid, 0);
+
+	StartTransactionCommand();
+	SPI_connect();
+	PushActiveSnapshot(GetTransactionSnapshot());
+	pgstat_report_activity(STATE_RUNNING, TAG ": resetting position");
+
+	pgrn_crash_safer_statuses_set_prepare_pid(NULL,
+											  databaseOid,
+											  tableSpaceOid,
+											  MyProcPid);
+	before_shmem_exit(pgroonga_crash_safer_prepare_one_on_exit,
+					  databaseInfoDatum);
+
+	{
+		int result;
+
+		SetCurrentStatementStartTimestamp();
+		result = SPI_execute("SELECT proname "
+							 "  FROM pg_catalog.pg_proc "
+							 "  WHERE "
+							 "    proname = 'pgroonga_wal_set_applied_position'",
+							 true,
+							 0);
+		if (result != SPI_OK_SELECT)
+		{
+			ereport(FATAL,
+					(errmsg(TAG ": failed to detect "
+							"pgroonga_wal_set_applied_position(): "
+							"%u/%u: %d",
+							databaseOid,
+							tableSpaceOid,
+							result)));
+		}
+
+		if (SPI_processed > 0)
+		{
+			SetCurrentStatementStartTimestamp();
+			result = SPI_execute("SELECT pgroonga_wal_set_applied_position()",
+								 false,
+								 0);
+			if (result != SPI_OK_SELECT)
+			{
+				ereport(FATAL,
+						(errmsg(TAG ": failed to reset WAL applied positions "
+								"of all PGroonga indexes: "
+								"%u/%u: %d",
+								databaseOid,
+								tableSpaceOid,
+								result)));
+			}
+		}
+	}
+
+	PopActiveSnapshot();
+	SPI_finish();
+	CommitTransactionCommand();
+
+	pgstat_report_activity(STATE_IDLE, NULL);
+
+	proc_exit(0);
+}
+
 void
 pgroonga_crash_safer_reindex_one(Datum databaseInfoDatum)
 {
@@ -147,11 +222,11 @@ pgroonga_crash_safer_reindex_one(Datum databaseInfoDatum)
 	PushActiveSnapshot(GetTransactionSnapshot());
 	pgstat_report_activity(STATE_RUNNING, TAG ": reindexing");
 
-	pgrn_crash_safer_statuses_set_reindex_pid(NULL,
+	pgrn_crash_safer_statuses_set_prepare_pid(NULL,
 											  databaseOid,
 											  tableSpaceOid,
 											  MyProcPid);
-	before_shmem_exit(pgroonga_crash_safer_reindex_one_on_exit,
+	before_shmem_exit(pgroonga_crash_safer_prepare_one_on_exit,
 					  databaseInfoDatum);
 
 	{
@@ -290,6 +365,8 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 	char pgrnDatabasePath[MAXPGPATH];
 	bool pgrnDatabasePathExist;
 	bool needReindex = false;
+	/* Only on the primary. */
+	bool needResetPosition = !RecoveryInProgress();
 	grn_ctx ctx;
 	grn_obj *db;
 	HTAB *statuses;
@@ -377,14 +454,15 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 	}
 	pfree(databasePath);
 
-	if (needReindex)
+	if (needReindex || needResetPosition)
 	{
 		BackgroundWorker worker = {0};
 		BackgroundWorkerHandle *handle;
 
 		snprintf(worker.bgw_name,
 				 BGW_MAXLEN,
-				 TAG ": reindex: %u/%u",
+				 TAG ": prepare: %s: %u/%u",
+				 needReindex ? "reindex" : "reset-position",
 				 databaseOid,
 				 tableSpaceOid);
 		snprintf(worker.bgw_type, BGW_MAXLEN, "%s", worker.bgw_name);
@@ -396,9 +474,10 @@ pgroonga_crash_safer_flush_one(Datum databaseInfoDatum)
 		snprintf(worker.bgw_library_name,
 				 BGW_MAXLEN,
 				 "%s", PGroongaCrashSaferLibraryName);
-		snprintf(worker.bgw_function_name,
-				 BGW_MAXLEN,
-				 "pgroonga_crash_safer_reindex_one");
+		snprintf(worker.bgw_function_name, BGW_MAXLEN,
+				 needReindex ?
+				 "pgroonga_crash_safer_reindex_one" :
+				 "pgroonga_crash_safer_reset_position_one");
 		worker.bgw_main_arg = databaseInfoDatum;
 		worker.bgw_notify_pid = MyProcPid;
 		if (RegisterDynamicBackgroundWorker(&worker, &handle))
